var _ = require('lodash');
var async = require('async');
var RSVP = require('rsvp');

var utils = require('../utils');

function getTree(params, cb) {
  function getTreeFromData(data) {
    return _.map(data, function(instance) {
      var children = [];
      if (instance.c && instance.c.length > 0) {
        children = getTreeFromData(instance.c);
      }
      return {
        instance: params.idMethods.fromId(instance.i),
        children: children
      };
    });
  }

  function fetchTreeAttributes(treeData, cb) {
    async.map(treeData, function(item, cb) {
      item.instance.fetch(params.valueOptions.attributes, params.valueOptions)
        .then(function() {
          if (item.children) {
            return fetchTreeAttributes(item.children, cb);
          }
          cb();
        }).catch(cb);
    }, cb);
  }

  function getUserTree(treeItems) {
    return _.map(treeItems, function(item) {
      var children = {};
      children[params.childrenAttribute] = getUserTree(item.children);
      return _.extend({},
        item.instance.get(params.valueOptions.attributes, params.valueOptions),
        children);
    });
  }

  return function(err, data) {
    if (err) {return cb(err);}
    var treeData = getTreeFromData(data);

    fetchTreeAttributes(treeData, function(err) {
      if (err) {return cb(err);}
      // Decide the format to return the results in.
      if (params.valueOptions.user) {
        // Return user-formatted data.
        cb(null, getUserTree(treeData));
      } else {
        // Return the instance objects if not in user mode.
        cb(null, treeData);
      }
    });
  };
}

function eachChild(instances, func) {
  return RSVP.all(instances, function(item) {
    return RSVP.all([
      func(item.instance),
      eachChild(item.children, func)
    ]);
  });
}

function getDescendants(instance, params) {
  // need:
  /*
instance, attribute, store, attributeName
type: redisCollectionType
  */
  var model = instance.model;
  var childModel = model.orm.models[params.type];
  var key = [
    model._props.keyspace,
    utils.serializeId(instance.id), attributeName
  ].join(';');

  var params = _.extend(options.tree, {
    rootKey: key,
    childKeyspace: childModel._props.keyspace,
    idMethods: idMethods,
    type: type
  });

  options.store.fetchTree(params, getTree({
    childModel: childModel,
    valueOptions: valueOptions,
    idMethods: idMethods,
    childrenAttribute: options.tree.childrenAttribute
  }, cb));
}

function syncDescendants(instance, value, params) {
  return getDescendants(instance, params.attribute)
    .then(function(instances) {
      return eachChild(instances, function(instance) {
        var data = {};
        data[params.link] = value;
        return instance.save(data);
      });
    });
}

function syncChildren(result, options) {
    var instanceDataList = [];
    var treeLink;

    if (result.action) {
      var data = {};
      var value = result.action === 'add' ? this : null;

      // Link HasMany children.
      if (options.link) {
        data[options.link] = value;
      }

      // Link immediate tree children.
      if (options.tree && options.tree.link) {
        treeLink = options.tree.link;
        data[treeLink] = value;
      }

      // Link tree descendants.
      if (options._treeLink) {
        treeLink = options._treeLink;
        data[treeLink] = value;
      }

      _.each(result.instances, function(instance) {
        instanceDataList.push({instance: instance, data: data});
      });
    }

    return RSVP.map(instanceDataList, function(instanceData) {
      if (_.isEmpty(instanceData.data)) {return RSVP.resolve();}
      return instanceData.instance
        .save(instanceData.data)
        // Traverse through instance children and set the tree link to value.
        .then(function() {
          if (treeLink) {
            return syncDescendants(instanceData.instance, value, {
              link: treeLink, attribute: attribute
            });
          }
        });
    });
}

function createRedisAssociation(type) {
  return function(options) {
    function executeCommands(name, value, attrOptions, cb) {
      var self = this;
      // Get the association redis key.
      var key = [this.model._props.keyspace, utils.serializeId(this.id), name]
        .join(';');
      var models = this.model.orm.models;
      var idMethods = utils.getIdMethods(options, models);

      // Get options about how to save/fetch instances in the association.
      var valueOptions = value || attrOptions.attributes[name];

      if (valueOptions && valueOptions.tree) {
        // A tree fetch was requested.
        if (!options.type) {
          throw new Error('related model type required for trees');
        }
        if (!options.tree.childrenAttribute) {
          throw new Error('childrenAttribute required for trees');
        }

        var childModel = models[options.type];
        var params = _.extend(options.tree, {
          rootKey: key,
          childKeyspace: childModel._props.keyspace,
          idMethods: idMethods,
          type: type
        });

        options.store.fetchTree(params, getTree({
          childModel: childModel,
          valueOptions: valueOptions,
          idMethods: idMethods,
          childrenAttribute: options.tree.childrenAttribute
        }, cb));
      } else {
        options.store.execute(type, idMethods, {
          key: key, query: valueOptions
        }, function(err, result) {
          if (err) {return cb(err);}
          syncChildren.call(self, result, options).then(function() {
            cb(null, result.data);
          }).catch(cb);
        });
      }
    }

    return {
      initialize: function() {
        // Set link for tree discovery.
        if (options.tree) {
          var targetModel = this.orm.models[options.type];
          var target = targetModel._attributes[options.tree.childrenAttribute];
          target._treeLink = options.tree.link;
        }
      },
      virtual: true,
      fetch: executeCommands,
      save: executeCommands,
      fetchTree: function() {

      }
    };
  };
}

exports.RedisList = createRedisAssociation('list');
exports.RedisSet = createRedisAssociation('set');
exports.RedisZSet = createRedisAssociation('zset');
