var RSVP = require('rsvp');
var redis = require('redis');
var chai = require('chai');

var datastored = require('../../..');
var redisDatastores = require('../../../lib/datastores/redis');
var testUtils = require('../../test_utils');

var expect = chai.expect;

describe('Redis associaitons >', function() {

  before(function() {
    this.env = testUtils.createTestEnv();
    this.client = redis.createClient();
    this.hashStore = new redisDatastores.RedisHashStore(this.client);
    this.store = new redisDatastores.RedisAssociationStore(this.client);

    this.ParentModel = this.env.createWithAttributes('ParentModel', {
      children: datastored.RedisList({
        store: this.store,
        type: 'Child3Model',
        link: 'parent'
      })
    });

    this.Child3Model = this.env.createWithAttributes('Child3Model', {
      parent: datastored.HasOne({
        hashStores: [this.hashStore],
        type: 'ParentModel'
      })
    });
  });

  beforeEach(function(done) {
    this.store.reset(done);
  });

  beforeEach(function(done) {
    this.hashStore.reset(done);
  });

  beforeEach(function() {
    var self = this;
    return testUtils.nextTick(function() {
      return RSVP.all([
        self.ParentModel.create().then(function(instance) {
          self.parent = instance;
        }),
        self.env.BasicModel.create().then(function(instance) {
          self.child1 = instance;
        }),
        self.env.BasicModel.create().then(function(instance) {
          self.child2 = instance;
        }),
        self.Child3Model.create().then(function(instance) {
          self.child3 = instance;
        })
      ]);
    }).then(function() {
      return self.parent.fetch({children: ['lpush', [
        self.child1, self.child2
      ]]});
    });
  });

  describe('RedisList', function() {

    it('should use lpush correctly', function() {
      return this.parent.fetch({
        children: ['lpush', [this.parent, this.parent]]
      }).then(function(data) {
        data.should.eq(4);
      });
    });

    it('should use lrange correctly', function() {
      var self = this;
      var parent = this.parent;
      return this.parent.fetch({children: ['lrange', [0, 4]]})
        .then(function(data) {
          testUtils.assertEqualInstances(data[0], self.child2);
          testUtils.assertEqualInstances(data[1], self.child1);
        });
    });

    it('should save correctly', function() {
      return this.parent.save({
        children: ['lpush', [this.parent, this.parent]]
      });
    });

    describe('1:n links', function() {

      it('should sync local changes to target instances', function() {
        add(parent, 'children', child)
        child.parent.should.eq(parent)
        child.parent.foo.eq(1)
        remove(parent, 'children', child)
        child.parent.should.eq(null)
      });

      it('should sync target changes to local instances', function() {
        save(child, 'parent', parent1)
        parent1.children.shouldInclude(child);
        child.parent.foo.eq('parent1value')
        save(child, 'parent', parent2)
        parent1.children.shouldNotInclude(child);
        parent2.children.shouldInclude(child);
        child.parent.foo.eq('parent2value')
        save(child, 'parent', null)
        parent2.children.shouldNotInclude(child);
      });
    });

    describe('n:n links', function() {

      // parent1 has link

      it('should sync local changes to target instances', function() {
        add(parent1, 'children', parent2)
        parent2.children.should.include(parent1)
        remove(parent1, 'children', parent2)
        parent2.children.should.notInclude(parent1)
      });

      it('should sync target changes to local instances', function() {
        add(parent2, 'children', parent1)
        parent1.children.should.include(parent2)
        remove(parent2, 'children', parent1)
        parent1.children.should.notInclude(parent2)
      });
    });

    describe('trees', function() {

      before(function() {
        this.ChildModel = this.env.createWithAttributes('ChildModel', {
          children: datastored.RedisList({
            store: this.store,
            type: 'ChildModel'
          }),
          root: datastored.HasOne({
            hashStores: [this.hashStore],
            type: 'AncestorModel'
          }),
          foo: datastored.Integer({hashStores: [this.hashStore]}),
          bar: datastored.Integer({hashStores: [this.hashStore]}),
          baz: datastored.Integer({hashStores: [this.hashStore]})
        });

        this.AncestorModel = this.env.createWithAttributes('AncestorModel', {
          descendants: datastored.RedisList({
            store: this.store,
            type: 'ChildModel',
            tree: {
              childrenAttribute: 'children',
              link: 'root'
            }
          }),
          foo: datastored.Integer({hashStores: [this.hashStore]})
        });
      });

      beforeEach(function() {
        // Save a test structure.
        var self = this;
        var child1, child2, child3;
        return this.ChildModel.create({foo: 1, bar: 2, baz: 3}).then(function(child) {
	  child1 = child;
          return self.ChildModel.create({foo: 4, bar: 5, baz: 6});
        }).then(function(child) {
          child2 = child;
          return self.ChildModel.create({foo: 7});
        }).then(function(child) {
          child3 = child;
          return self.ChildModel.create({foo: 4, children: ['rpush', [child1, child2]]});
        }).then(function(child) {
          return self.AncestorModel.create({foo: 5, descendants: ['rpush', [child]]});
        }).then(function(parent) {
          self.parent = parent;
        });
      });

      it('should be fetched correctly with user transforms', function() {
        return this.parent.fetch({descendants: {tree: true, attributes: {
          foo: true, bar: true
        }, user: true}}).then(function(descendants) {
          testUtils.assertEqualResults(descendants, [{
            children: [{
              id: true,
              foo: 1,
              bar: 2,
              children: []
            }, {
              id: true,
              foo: 4,
              bar: 5,
              children: []
            }]
          }]);
        });
      });

      it('should be fetched correctly without user transforms', function() {
        return this.parent.fetch({descendants: {tree: true, attributes: {
          foo: true, bar: true
        }}}).then(function(descendants) {
          var attributes = {foo: true, bar: true, baz: true};
          descendants[0].instance.get('foo').should.eq(4);
          descendants[0].children[0].instance.get(attributes, {ids: false})
            .should.deep.eq({foo: 1, bar: 2});
          descendants[0].children[1].instance.get(attributes, {ids: false})
            .should.deep.eq({foo: 4, bar: 5});
        });
      });

      describe('links', function() {

        it('should sync root references', function() {
          root.save({foo: '1'})

          child1.add(child2)
          parent.add(child1)

          root.add(parent)

          parent.get('root').should.eq(root)
          parent.get('root').get('foo').should.eq('1')
          child1.get('root').should.equal(root)
          child1.get('root').get('foo').should.eq('1')
          child2.get('root').should.equal(root)
          child2.get('root').get('foo').should.eq('1')

          root.remove(parent)
          newRoot.add(parent)

          parent.get('root').should.eq(newRoot)
          parent.get('root').get('foo').should.eq('2')
          child1.get('root').should.equal(newRoot)
          child1.get('root').get('foo').should.eq('2')
          child2.get('root').should.equal(newRoot)
          child2.get('root').get('foo').should.eq('2')

          // Test that joined attributes are synced.
          newRoot.change('joined attribute')
          // children should have new joined attributes

          // should sync if references are changed
          parent.set('root', root);
          // children should change root and should have new joined attrs
        });
      });
    });
  });
});
